# -*- coding: utf-8 -*-

from copy import deepcopy

from .request import SnowRequest
from .url_builder import URLBuilder


class Resource(object):
    """Creates a new :class:`Resource` object

    Resources provides a natural way of interfacing with ServiceNow APIs.

    :param base_path: Base path
    :param api_path: API path
    :param chunk_size: Response stream parser chunk size (in bytes)
    :param \*\*kwargs: Arguments to pass along to :class:`Request`
    """

    def __init__(self, base_url=None, base_path=None, api_path=None, parameters=None, **kwargs):

        self._base_url = base_url
        self._base_path = base_path
        self._api_path = api_path
        self._url_builder = URLBuilder(base_url, base_path, api_path)

        self.kwargs = kwargs

        self.parameters = deepcopy(parameters)

    def __repr__(self):
        return '<%s [%s]>' % (self.__class__.__name__, self.path)

    @property
    def path(self):
        return "%s" % self._base_path + self._api_path

    @property
    def _request(self):
        parameters = deepcopy(self.parameters)

        return SnowRequest(url_builder=self._url_builder, parameters=parameters, **self.kwargs)

    def get(self, query, limit=None, offset=None, fields=list()):
        """Queries the API resource

        :param query: Dictionary, string or :class:`QueryBuilder` object
        :param limit: (optional) Limits the number of records returned
        :param fields: (optional) List of fields to include in the response created_on in descending order.
        :param offset: (optional) Number of records to skip before returning records
        :return: :class:`Response` object
        """

        return self._request.get(query, limit, offset, fields)

    def create(self, payload):
        """Creates a new record in the API resource

        :param payload: Dictionary containing key-value fields of the new record
        :return: Dictionary of the inserted record
        """

        return self._request.create(payload)

    def update(self, query, payload):
        """Updates a record in the API resource

        :param query: Dictionary, string or :class:`QueryBuilder` object
        :param payload: Dictionary containing key-value fields of the record to be updated
        :return: Dictionary of the updated record
        """

        return self._request.update(query, payload)

    def delete(self, query):
        """Deletes matching record

        :param query: Dictionary, string or :class:`QueryBuilder` object
        :return: Dictionary containing information about deletion result
        """

        return self._request.delete(query)

    def custom(self, method, path_append=None, headers=None, **kwargs):
        """Creates a custom request

        :param method: HTTP method to use
        :param path_append: (optional) relative to :attr:`api_path`
        :param headers: (optional) Dictionary of headers to add or override
        :param kwargs: kwargs to pass along to :class:`requests.Request`
        :return: :class:`Response` object
        """

        return self._request.custom(method, path_append=path_append, headers=headers, **kwargs)



# -*- coding: utf-8 -*-

# Public API
from .client import Client
from .oauth_client import OAuthClient
from .query_builder import QueryBuilder
from .resource import Resource

__author__ = "Robert Wikman <rbw@vault13.org>"
__version__ = "0.6.1"



# -*- coding: utf-8 -*-

import itertools
import json
import os
import six
import ntpath
import warnings

from pysnow.query_builder import QueryBuilder

from .legacy_exceptions import (NoRequestExecuted,
                                MultipleResults,
                                NoResults,
                                InvalidUsage,
                                UnexpectedResponse,
                                MissingResult)


class LegacyRequest(object):
    base_path = "api/now"

    def __init__(self, method, table, **kwargs):
        """Takes arguments used to perform a HTTP request

        :param method: HTTP request method
        :param table: table to operate on
        """
        self.method = method
        self.table = table
        self.url_link = None  # Updated when a linked request is iterated on
        self.base_url = kwargs.pop('base_url')
        self.request_params = kwargs.pop('request_params')
        self.raise_on_empty = kwargs.pop('raise_on_empty')
        self.session = kwargs.pop('session')
        self._last_response = None

        if method in ('GET', 'DELETE'):
            self.query = kwargs.pop('query')

    @property
    def last_response(self):
        """Return _last_response after making sure an inner `requests.request` has been performed

        :raise:
            :NoRequestExecuted: If no request has been executed
        :return: last response
        """
        if self._last_response is None:
            raise NoRequestExecuted("%s hasn't been executed" % self)
        return self._last_response

    @last_response.setter
    def last_response(self, response):
        """ Sets last_response property
        :param response: `requests.request` response
        """
        self._last_response = response

    @property
    def count(self):
        """ Returns the number of records the query would yield"""
        self.request_params.update({'sysparm_count': True})
        response = self.session.get(self._get_stats_url(),
                                    params=self._get_formatted_query(fields=list(),
                                                                     limit=None,
                                                                     order_by=list(),
                                                                     offset=None))

        content = self._get_content(response)

        return int(content['stats']['count'])

    @property
    def status_code(self):
        """Return last_response.status_code after making sure an inner `requests.request` has been performed

        :return: status_code of last_response
        """
        return self.last_response.status_code

    def _all_inner(self, fields, limit, order_by, offset):
        """Yields all records for the query and follows links if present on the response after validating

        :return: List of records with content
        """
        response = self.session.get(self._get_table_url(),
                                    params=self._get_formatted_query(fields, limit, order_by, offset))

        yield self._get_content(response)
        while 'next' in response.links:
            self.url_link = response.links['next']['url']
            response = self.session.get(self.url_link)
            yield self._get_content(response)

    def get_all(self, fields=list(), limit=None, order_by=list(), offset=None):
        """DEPRECATED - see get_multiple()"""
        warnings.warn("get_all() is deprecated, please use get_multiple() instead", DeprecationWarning)
        return self.get_multiple(fields, limit, order_by, offset)

    def get_multiple(self, fields=list(), limit=None, order_by=list(), offset=None):
        """Wrapper method that takes whatever was returned by the _all_inner() generators and chains it in one result

        The response can be sorted by passing a list of fields to order_by.

        Example:
        get_multiple(order_by=['category', '-created_on']) would sort the category field in ascending order,
        with a secondary sort by created_on in descending order.

        :param fields: List of fields to return in the result
        :param limit: Limits the number of records returned
        :param order_by: Sort response based on certain fields
        :param offset: A number of records to skip before returning records (for pagination)
        :return: Iterable chain object
        """
        return itertools.chain.from_iterable(self._all_inner(fields, limit, order_by, offset))

    def get_one(self, fields=list()):
        """Convenience function for queries returning only one result. Validates response before returning.

        :param fields: List of fields to return in the result
        :raise:
            :MultipleResults: if more than one match is found
        :return: Record content
        """
        response = self.session.get(self._get_table_url(),
                                    params=self._get_formatted_query(fields, limit=None, order_by=list(), offset=None))

        content = self._get_content(response)
        l = len(content)
        if l > 1:
            raise MultipleResults('Multiple results for get_one()')

        if len(content) == 0:
            return {}

        return content[0]

    def insert(self, payload):
        """Inserts a new record with the payload passed as an argument

        :param payload: The record to create (dict)
        :return: Created record
        """
        response = self.session.post(self._get_table_url(), data=json.dumps(payload))
        return self._get_content(response)

    def delete(self):
        """Deletes the queried record and returns response content after response validation

        :raise:
            :NoResults: if query returned no results
            :NotImplementedError: if query returned more than one result (currently not supported)
        :return: Delete response content (Generally always {'Success': True})
        """
        try:
            result = self.get_one()
            if 'sys_id' not in result:
                raise NoResults()
        except MultipleResults:
            raise MultipleResults("Deletion of multiple records is not supported")
        except NoResults as e:
            e.args = ('Cannot delete a non-existing record',)
            raise

        response = self.session.delete(self._get_table_url(sys_id=result['sys_id']))
        return self._get_content(response)

    def update(self, payload):
        """Updates the queried record with `payload` and returns the updated record after validating the response

        :param payload: Payload to update the record with
        :raise:
            :NoResults: if query returned no results
            :MultipleResults: if query returned more than one result (currently not supported)
        :return: The updated record
        """
        try:
            result = self.get_one()
            if 'sys_id' not in result:
                raise NoResults()
        except MultipleResults:
            raise MultipleResults("Update of multiple records is not supported")
        except NoResults as e:
            e.args = ('Cannot update a non-existing record',)
            raise

        if not isinstance(payload, dict):
            raise InvalidUsage("Update payload must be of type dict")

        response = self.session.put(self._get_table_url(sys_id=result['sys_id']), data=json.dumps(payload))
        return self._get_content(response)

    def clone(self, reset_fields=list()):
        """Clones the queried record

        :param reset_fields: Fields to reset
        :raise:
            :NoResults: if query returned no results
            :MultipleResults: if query returned more than one result (currently not supported)
            :UnexpectedResponse: informs the user about what likely went wrong
        :return: The cloned record
        """

        if not isinstance(reset_fields, list):
            raise InvalidUsage("reset_fields must be a `list` of fields")

        try:
            response = self.get_one()
            if 'sys_id' not in response:
                raise NoResults()
        except MultipleResults:
            raise MultipleResults('Cloning multiple records is not supported')
        except NoResults as e:
            e.args = ('Cannot clone a non-existing record',)
            raise

        payload = {}

        # Iterate over fields in the result
        for field in response:
            # Ignore fields in reset_fields
            if field in reset_fields:
                continue

            item = response[field]
            # Check if the item is of type dict and has a sys_id ref (value)
            if isinstance(item, dict) and 'value' in item:
                payload[field] = item['value']
            else:
                payload[field] = item

        try:
            return self.insert(payload)
        except UnexpectedResponse as e:
            if e.status_code == 403:
                # User likely attempted to clone a record without resetting a unique field
                e.args = ('Unable to create clone. Make sure unique fields has been reset.',)
            raise

    def attach(self, file):
        """Attaches the queried record with `file` and returns the response after validating the response

        :param file: File to attach to the record
        :raise:
            :NoResults: if query returned no results
            :MultipleResults: if query returned more than one result (currently not supported)
        :return: The attachment record metadata
        """
        try:
            result = self.get_one()
            if 'sys_id' not in result:
                raise NoResults()
        except MultipleResults:
            raise MultipleResults('Attaching a file to multiple records is not supported')
        except NoResults:
            raise NoResults('Attempted to attach file to a non-existing record')

        if not os.path.isfile(file):
            raise InvalidUsage("Attachment '%s' must be an existing regular file" % file)

        response = self.session.post(
            self._get_attachment_url('upload'),
            data={
                'table_name': self.table,
                'table_sys_id': result['sys_id'],
                'file_name': ntpath.basename(file)
            },
            files={'file': open(file, 'rb')},
            headers={'content-type': None}  # Temporarily override header
        )
        return self._get_content(response)

    def _get_content(self, response):
        """Checks for errors in the response. Returns response content, in bytes.

        :param response: response object
        :raise:
            :UnexpectedResponse: if the server responded with an unexpected response
        :return: ServiceNow response content
        """
        method = response.request.method
        self.last_response = response

        server_error = {
            'summary': None,
            'details': None
        }

        try:
            content_json = response.json()
            if 'error' in content_json:
                e = content_json['error']
                if 'message' in e:
                    server_error['summary'] = e['message']
                if 'detail' in e:
                    server_error['details'] = e['detail']
        except ValueError:
            content_json = {}

        if method == 'DELETE':
            # Make sure the delete operation returned the expected response
            if response.status_code == 204:
                return {'success': True}
            else:
                raise UnexpectedResponse(
                    204, response.status_code, method,
                    server_error['summary'], server_error['details']
                )
        # Make sure the POST operation returned the expected response
        elif method == 'POST' and response.status_code != 201:
            raise UnexpectedResponse(
                201, response.status_code, method,
                server_error['summary'], server_error['details']
            )
        # It seems that Helsinki and later returns status 200 instead of 404 on empty result sets
        if ('result' in content_json and len(content_json['result']) == 0) or response.status_code == 404:
            if self.raise_on_empty is True:
                raise NoResults('Query yielded no results')
        elif 'error' in content_json:
            raise UnexpectedResponse(
                200, response.status_code, method,
                server_error['summary'], server_error['details']
            )

        if 'result' not in content_json:
            raise MissingResult("The request was successful but the content didn't contain the expected 'result'")

        return content_json['result']

    def _get_table_url(self, **kwargs):
        return self._get_url('table', item=self.table, **kwargs)

    def _get_attachment_url(self, action):
        return self._get_url('attachment', item=action)

    def _get_stats_url(self):
        return self._get_url('stats', item=self.table)

    def _get_url(self, resource, item, sys_id=None):
        """Takes table and sys_id (if present), and returns a URL

        :param resource: API resource
        :param item: API resource item
        :param sys_id: Record sys_id
        :return: url string
        """

        url_str = '%(base_url)s/%(base_path)s/%(resource)s/%(item)s' % (
            {
                'base_url': self.base_url,
                'base_path': self.base_path,
                'resource': resource,
                'item': item
            }
        )

        if sys_id:
            return "%s/%s" % (url_str, sys_id)

        return url_str

    def _get_formatted_query(self, fields, limit, order_by, offset):
        """
        Converts the query to a ServiceNow-interpretable format
        :return: ServiceNow query
        """

        if not isinstance(order_by, list):
            raise InvalidUsage("Argument order_by should be a `list` of fields")

        if not isinstance(fields, list):
            raise InvalidUsage("Argument fields should be a `list` of fields")

        if isinstance(self.query, QueryBuilder):
            sysparm_query = str(self.query)
        elif isinstance(self.query, dict):  # Dict-type query
            sysparm_query = '^'.join(['%s=%s' % (k, v) for k, v in six.iteritems(self.query)])
        elif isinstance(self.query, str):  # String-type query
            sysparm_query = self.query
        else:
            raise InvalidUsage("Query must be instance of %s, %s or %s" % (QueryBuilder, str, dict))

        for field in order_by:
            if field[0] == '-':
                sysparm_query += "^ORDERBYDESC%s" % field[1:]
            else:
                sysparm_query += "^ORDERBY%s" % field

        params = {'sysparm_query': sysparm_query}
        params.update(self.request_params)

        if limit is not None:
            params.update({'sysparm_limit': limit, 'sysparm_suppress_pagination_header': True})

        if offset is not None:
            params.update({'sysparm_offset': offset})

        if len(fields) > 0:
            params.update({'sysparm_fields': ",".join(fields)})

        return params




# -*- coding: utf-8 -*-

import six

from .query_builder import QueryBuilder

from .exceptions import InvalidUsage


class ParamsBuilder(object):
    """Provides an interface for setting / getting common ServiceNow sysparms."""

    def __init__(self):
        self._custom_params = {}

        self._sysparms = {
            'sysparm_query': '',
            'sysparm_limit': 10000,
            'sysparm_offset': None,
            'sysparm_display_value': False,
            'sysparm_suppress_pagination_header': False,
            'sysparm_exclude_reference_link': False,
            'sysparm_view': '',
            'sysparm_fields': []
        }

    @staticmethod
    def stringify_query(query):
        """Stringifies the query (dict or QueryBuilder) into a ServiceNow-compatible format

        :return:
            - ServiceNow-compatible string-type query
        """

        if isinstance(query, QueryBuilder):
            # Get string-representation of the passed :class:`pysnow.QueryBuilder` object
            return str(query)
        elif isinstance(query, dict):
            # Dict-type query
            return '^'.join(['%s=%s' % (k, v) for k, v in six.iteritems(query)])
        elif isinstance(query, str):
            # Regular string-type query
            return query
        else:
            raise InvalidUsage('Query must be of type string, dict or a QueryBuilder object')

    def add_custom(self, params):
        """Adds new custom parameter after making sure it's of type dict.

        :param params: Dictionary containing one or more parameters
        """

        if isinstance(params, dict) is False:
            raise InvalidUsage("custom parameters must be of type `dict`")

        self._custom_params.update(params)

    @property
    def custom_params(self):
        """Returns a dictionary of added custom parameters"""
        return self._custom_params

    @property
    def display_value(self):
        """Maps to `sysparm_display_value`"""
        return self._sysparms['sysparm_display_value']

    @display_value.setter
    def display_value(self, value):
        """Sets `sysparm_display_value`

        :param value:  Bool or 'all'
        """

        if not (isinstance(value, bool) or value == 'all'):
            raise InvalidUsage("Display value can be of type bool or value 'all'")

        self._sysparms['sysparm_display_value'] = value

    @property
    def query(self):
        """Maps to `sysparm_query`"""
        return self._sysparms['sysparm_query']

    @query.setter
    def query(self, query):
        """Validates, stringifies and sets `sysparm_query`

        :param query: String, dict or QueryBuilder
        """

        self._sysparms['sysparm_query'] = self.stringify_query(query)

    @property
    def limit(self):
        """Maps to `sysparm_limit`"""
        return self._sysparms['sysparm_limit']

    @limit.setter
    def limit(self, limit):
        """Sets `sysparm_limit`

        :param limit: Size limit (int)
        """

        if not isinstance(limit, int) or isinstance(limit, bool):
            raise InvalidUsage("limit size must be of type integer")

        self._sysparms['sysparm_limit'] = limit

    @property
    def offset(self):
        """Maps to `sysparm_offset`"""
        return self._sysparms['sysparm_offset']

    @offset.setter
    def offset(self, offset):
        """Sets `sysparm_offset`, usually used to accomplish pagination

        :param offset: Number of records to skip before fetching records
        :raise:
            :InvalidUsage: if offset is of an unexpected type
        """

        if not isinstance(offset, int) or isinstance(offset, bool):
            raise InvalidUsage('Offset must be an integer')

        self._sysparms['sysparm_offset'] = offset

    @property
    def fields(self):
        """Maps to `sysparm_fields`"""
        return self._sysparms['sysparm_fields']

    @fields.setter
    def fields(self, fields):
        """Sets `sysparm_fields` after joining the given list of `fields`

        :param fields: List of fields to include in the response
        :raise:
            :InvalidUsage: if fields is of an unexpected type
        """

        if not isinstance(fields, list):
            raise InvalidUsage('fields must be of type `list`')

        self._sysparms['sysparm_fields'] = ",".join(fields)

    @property
    def exclude_reference_link(self):
        """Maps to `sysparm_exclude_reference_link`"""
        return self._sysparms['sysparm_exclude_reference_link']

    @exclude_reference_link.setter
    def exclude_reference_link(self, exclude):
        """Sets `sysparm_exclude_reference_link` to a bool value

        :param exclude: bool
        """
        if not isinstance(exclude, bool):
            raise InvalidUsage('exclude_reference_link must be of type bool')

        self._sysparms['sysparm_exclude_reference_link'] = exclude

    @property
    def suppress_pagination_header(self):
        """Maps to `sysparm_suppress_pagination_header`"""
        return self._sysparms['sysparm_suppress_pagination_header']

    @suppress_pagination_header.setter
    def suppress_pagination_header(self, suppress):
        """Enables or disables pagination header by setting `sysparm_suppress_pagination_header`

        :param suppress: bool
        """
        if not isinstance(suppress, bool):
            raise InvalidUsage('suppress_pagination_header must be of type bool')

        self._sysparms['sysparm_suppress_pagination_header'] = suppress

    def as_dict(self):
        """Constructs query params compatible with :class:`requests.Request`

        :return:
            - Dictionary containing query parameters
        """

        sysparms = self._sysparms
        sysparms.update(self._custom_params)

        return sysparms



# -*- coding: utf-8 -*-

import ijson

from ijson.common import ObjectBuilder

from itertools import chain

from .exceptions import (ResponseError,
                         NoResults,
                         MultipleResults,
                         MissingResult)


class Response(object):
    """Takes a :class:`requests.Response` object and performs deserialization and validation.

    :param response: :class:`request.Response` object
    :param chunk_size: Read and return up to this size (in bytes) in the stream parser
    """

    def __init__(self, response, chunk_size=2048):
        self._response = response
        self._chunk_size = chunk_size
        self._count = 0

    @property
    def count(self):
        return self._count

    @count.setter
    def count(self, count):
        if not isinstance(count, int) or isinstance(count, bool):
            raise TypeError("Count must be an integer")

        self._count = count

    def __repr__(self):
        return '<%s [%d - %s]>' % (self.__class__.__name__, self._response.status_code, self._response.request.method)

    def _parse_response(self):
        """Looks for `result.item` (array), `result` (object) and `error` (object) keys and parses
        the raw response content (stream of bytes)

        :raise:
            - ResponseError: If there's an error in the response
            - MissingResult: If no result nor error was found
        """

        has_result_single = False
        has_result_many = False
        has_error = False

        for prefix, event, value in ijson.parse(self._response.raw, buf_size=self._chunk_size):
            if (prefix, event) == ('error', 'start_map'):
                # Matched ServiceNow `error` object at the root
                has_error = True
                builder = ObjectBuilder()
            elif prefix == 'result' and event in ['start_map', 'start_array']:
                # Matched ServiceNow `result`
                builder = ObjectBuilder()
                if event == 'start_map':  # Matched object
                    has_result_single = True
                elif event == 'start_array':  # Matched array
                    has_result_many = True

            if has_result_many:
                # Build the result
                if (prefix, event) == ('result.item', 'end_map'):
                    # Reached end of object. Set count and yield
                    builder.event(event, value)
                    self.count += 1
                    yield getattr(builder, 'value')
                elif prefix.startswith('result.item'):
                    # Build the result object
                    builder.event(event, value)
            elif has_result_single:
                if (prefix, event) == ('result', 'end_map'):
                    # Reached end of the result object. Set count and yield.
                    builder.event(event, value)
                    self.count += 1
                    yield getattr(builder, 'value')
                elif prefix.startswith('result'):
                    # Build the error object
                    builder.event(event, value)
            elif has_error:
                if (prefix, event) == ('error', 'end_map'):
                    # Reached end of the error object - raise ResponseError exception
                    raise ResponseError(getattr(builder, 'value'))
                elif prefix.startswith('error'):
                    # Build the error object
                    builder.event(event, value)

        if (has_result_single or has_result_many) and self.count == 0:  # Results empty
            yield {}

        if not (has_result_single or has_result_many or has_error):  # None of the expected keys were found
            raise MissingResult('The expected `result` key was missing in the response. Cannot continue')

    def _get_validated_response(self):
        """Validates response then calls :meth:`_parse_response` to yield content

        Immediately yields response content if request method is DELETE and code 204 (this response never
        contains a body).

        :raise:
            - HTTPError: if a non-200 response is encountered
        """

        response = self._response

        if response.request.method == 'DELETE' and response.status_code == 204:
            yield [{'status': 'record deleted'}]
        else:
            # Raise an HTTPError if we hit a non-200 status code
            response.raise_for_status()

            # Parse byte stream
            yield self._parse_response()

    def all(self):
        """Returns a chained generator response containing all matching records

        :return: Iterable response
        """

        return chain.from_iterable(self._get_validated_response())

    def first(self):
        """Return the first record or raise an exception if the result doesn't contain any data

        :return: Dictionary containing the first item in the response content
        :raise:
            - NoResults: If no results were found
        """

        content = next(self.all())

        if len(content) == 0:
            raise NoResults("No records found")

        return content


    def first_or_none(self):
        """Return the first record or None

        :return: Dictionary containing the first item or None
        """

        try:
            return self.first()
        except NoResults:
            return None

    def one(self):
        """Return exactly one result or raise an exception.

        :return: Dictionary containing the only item in the response content
        :raise:
            - MultipleResults: If more than one records are present in the content
            - NoResults: If no records are present in the content
        """

        r = self.all()
        result = next(r)

        if len(result) == 0:
            raise NoResults("No records found")

        try:
            next(r)
        except StopIteration:
            pass
        else:
            raise MultipleResults("Expected single-record result, got multiple")

        return result

    def one_or_none(self):
        """Return at most one result or raise an exception.

        :return: Dictionary containing the matching record or None
        :raise:
            - MultipleResults: If more than one records are present in the content
        """

        try:
            return self.one()
        except NoResults:
            return None



# -*- coding: utf-8 -*-

import json
import requests

from .response import Response
from .exceptions import InvalidUsage


class SnowRequest(object):
    """Creates a new :class:`SnowRequest` object.

    :param parameters: :class:`sysparms.Sysparms` object
    :param session: :class:`request.Session` object
    :param url_builder: :class:`url_builder.URLBuilder` object
    """

    def __init__(self, parameters=None, session=None, url_builder=None, chunk_size=None):
        self._parameters = parameters
        self._url_builder = url_builder
        self._session = session
        self._chunk_size = chunk_size or 4096

        self._url = url_builder.get_url()

    def _get_response(self, method, **kwargs):
        """Response wrapper - creates a :class:`requests.Response` object and passes along to :class:`pysnow.Response`
        for validation and parsing.

        :param args: args to pass along to _send()
        :param kwargs: kwargs to pass along to _send()
        :return: :class:`pysnow.Response` object
        """

        request = requests.Request(method, self._url, **kwargs)
        prepped = self._session.prepare_request(request)
        response = self._session.send(prepped,
                                      stream=True)

        response.raw.decode_content = True

        return Response(response, self._chunk_size)

    def get(self, query, limit=None, offset=None, fields=list()):
        """Fetches one or more records, exposes a public API of :class:`pysnow.Response`

        :param query: Dictionary, string or :class:`QueryBuilder` object
        :param limit: Limits the number of records returned
        :param fields: List of fields to include in the response
        created_on in descending order.
        :param offset: Number of records to skip before returning records
        :return: :class:`pysnow.Response` object
        """

        self._parameters.query = query

        if limit is not None:
            self._parameters.limit = limit

        if offset is not None:
            self._parameters.offset = offset

        if len(fields) > 0:
            self._parameters.fields = fields

        return self._get_response('GET', params=self._parameters.as_dict())

    def create(self, payload):
        """Creates a new record

        :param payload: Dictionary payload
        :return: Dictionary of the inserted record
        """

        return self._get_response('POST', data=json.dumps(payload)).one()

    def update(self, query, payload):
        """Updates a record

        :param query: Dictionary, string or :class:`QueryBuilder` object
        :param payload: Dictionary payload
        :return: Dictionary of the updated record
        """

        if not isinstance(payload, dict):
            raise InvalidUsage("Update payload must be of type dict")

        record = self.get(query).one()

        self._url = self._url_builder.get_appended_custom("/{0}".format(record['sys_id']))
        return self._get_response('PUT', data=json.dumps(payload)).one()

    def delete(self, query):
        """Deletes a record

        :param query: Dictionary, string or :class:`QueryBuilder` object
        :return: Dictionary containing status of the delete operation
        """

        record = self.get(query=query).one()

        self._url = self._url_builder.get_appended_custom("/{0}".format(record['sys_id']))
        return self._get_response('DELETE').one()

    def custom(self, method, path_append=None, headers=None, **kwargs):
        """Creates a custom request

        :param method: HTTP method
        :param path_append: (optional) append path to resource.api_path
        :param headers: (optional) Dictionary of headers to add or override
        :param kwargs: kwargs to pass along to :class:`requests.Request`
        :return: :class:`pysnow.Response` object
        """

        if headers:
            self._session.headers.update(headers)

        if path_append is not None:
            try:
                self._url = self._url_builder.get_appended_custom(path_append)
            except InvalidUsage:
                raise InvalidUsage("Argument 'path_append' must be a string in the following format: "
                                   "/path-to-append[/.../...]")

        return self._get_response(method, **kwargs)



# -*- coding: utf-8 -*-


class InvalidUsage(Exception):
    pass


class QueryTypeError(TypeError):
    pass


class QueryMissingField(Exception):
    pass


class QueryEmpty(Exception):
    pass


class QueryExpressionError(Exception):
    pass


class QueryMultipleExpressions(Exception):
    pass


class MissingResult(Exception):
    pass


class MissingToken(Exception):
    pass


class UnexpectedResponseFormat(Exception):
    pass


class ReportUnavailable(Exception):
    pass


class UnexpectedResponse(Exception):
    """Provides detailed information about a server error response

    :param code_expected: Expected HTTP status code
    :param code_actual: Actual HTTP status code
    :param http_method: HTTP method used
    :param error_summary: Summary of what went wrong
    :param error_details: Details about the error
    """

    def __init__(self, code_expected, code_actual, http_method, error_summary, error_details):
        if code_expected == code_actual:
            message = "Unexpected response on HTTP %s from server: %s" % (
                http_method,
                error_summary
            )
        else:
            message = "Unexpected HTTP %s response code. Expected %d, got %d" % (
                http_method,
                code_expected,
                code_actual
            )

        super(UnexpectedResponse, self).__init__(message)
        self.status_code = code_actual
        self.error_summary = error_summary
        self.error_details = error_details


class NoResults(Exception):
    pass


class MultipleResults(Exception):
    pass


class NoRequestExecuted(Exception):
    pass



# -*- coding: utf-8 -*-


######################
# Generic exceptions #
######################
class InvalidUsage(Exception):
    pass


#######################
# Response exceptions #
#######################
class ResponseError(Exception):
    message = '<empty>'
    detail = '<empty>'

    def __init__(self, error):
        if 'message' in error:
            self.message = error['message'] or self.message
        if 'detail' in error:
            self.detail = error['detail'] or self.detail

    def __str__(self):
        return 'Error in response. Message: %s, Details: %s' % (self.message, self.detail)


class MissingResult(Exception):
    pass


class UnexpectedResponseFormat(Exception):
    pass


class ReportUnavailable(Exception):
    pass


class NoResults(Exception):
    pass


class MultipleResults(Exception):
    pass


##########################
# OAuthClient exceptions #
##########################
class MissingToken(Exception):
    pass


class TokenCreateError(Exception):
    def __init__(self, error, description):
        message = "Error creating new user token"
        super(TokenCreateError, self).__init__(message)
        self.error = error
        self.description = description


############################
# Query builder exceptions #
############################
class QueryTypeError(TypeError):
    pass


class QueryMissingField(Exception):
    pass


class QueryEmpty(Exception):
    pass


class QueryExpressionError(Exception):
    pass


class QueryMultipleExpressions(Exception):
    pass



# -*- coding: utf-8 -*-

import warnings

from oauthlib.oauth2 import LegacyApplicationClient
from oauthlib.oauth2.rfc6749.errors import OAuth2Error
from requests_oauthlib import OAuth2Session

from .client import Client
from .exceptions import InvalidUsage, MissingToken, TokenCreateError

warnings.simplefilter("always", DeprecationWarning)


class OAuthClient(Client):
    """Pysnow `Client` with extras for oauth session and token handling.

    :param client_id: client_id from ServiceNow
    :param client_secret: client_secret from ServiceNow
    :param token_updater: function called when a token has been refreshed
    :param kwargs: kwargs passed along to :class:`pysnow.Client`
    """

    token = None

    def __init__(self, client_id=None, client_secret=None, token_updater=None, **kwargs):

        if not (client_secret and client_id):
            raise InvalidUsage('You must supply a client_id and client_secret')

        if kwargs.get('session') or kwargs.get('user'):
            warnings.warn('pysnow.OAuthClient manages sessions internally, '
                          'provided user / password credentials or sessions will be ignored.')

        # Forcibly set session, user and password
        kwargs['session'] = OAuth2Session(client=LegacyApplicationClient(client_id=client_id))
        kwargs['user'] = None
        kwargs['password'] = None

        super(OAuthClient, self).__init__(**kwargs)

        self.token_updater = token_updater
        self.client_id = client_id
        self.client_secret = client_secret

        self.token_url = "%s/oauth_token.do" % self.base_url

    def _get_oauth_session(self):
        """Creates a new OAuth session

        :return: OAuth2Session object
        """

        return OAuth2Session(
            client_id=self.client_id,
            token=self.token,
            token_updater=self.token_updater,
            auto_refresh_url=self.token_url,
            auto_refresh_kwargs={
                "client_id": self.client_id,
                "client_secret": self.client_secret
            })

    def set_token(self, token):
        """Sets token after validating

        :param token: dict containing the information required to create an OAuth2Session
        """

        if not token:
            self.token = None
            return

        expected_keys = set(("token_type", "refresh_token", "access_token", "scope", "expires_in", "expires_at"))
        if not isinstance(token, dict) or not expected_keys <= set(token):
            raise InvalidUsage("Token should contain a dictionary obtained using fetch_token()")

        self.token = token

    def _legacy_request(self, *args, **kwargs):
        """Makes sure token has been set, then calls parent to create a new :class:`pysnow.LegacyRequest` object

        :param args: args to pass along to _legacy_request()
        :param kwargs: kwargs to pass along to _legacy_request()
        :return: :class:`pysnow.LegacyRequest` object
        :raises:
            - MissingToken: If token hasn't been set
        """

        if isinstance(self.token, dict):
            self.session = self._get_oauth_session()
            return super(OAuthClient, self)._legacy_request(*args, **kwargs)

        raise MissingToken("You must set_token() before creating a legacy request with OAuthClient")

    def resource(self, api_path=None, base_path='/api/now', chunk_size=None):
        """Overrides :meth:`resource` provided by :class:`pysnow.Client` with extras for OAuth

        :param api_path: Path to the API to operate on
        :param base_path: (optional) Base path override
        :param chunk_size: Response stream parser chunk size (in bytes)
        :return: :class:`Resource` object
        :raises:
            - InvalidUsage: If a path fails validation
        """

        if isinstance(self.token, dict):
            self.session = self._get_oauth_session()
            return super(OAuthClient, self).resource(api_path, base_path)

        raise MissingToken("You must set_token() before creating a resource with OAuthClient")

    def generate_token(self, user, password):
        """Takes user and password credentials and generates a new token

        :param user: user
        :param password: password
        :return: dictionary containing token data
        :raises:
            - TokenCreateError: If there was an error generating the new token
        """

        try:
            return dict(self.session.fetch_token(token_url=self.token_url,
                                                 username=user,
                                                 password=password,
                                                 client_id=self.client_id,
                                                 client_secret=self.client_secret))
        except OAuth2Error as e:
            raise TokenCreateError(error=e.error, description=e.description)




# -*- coding: utf-8 -*-

import re
from .exceptions import InvalidUsage


class URLBuilder(object):
    def __init__(self, base_url, base_path, api_path):
        self._base_url = base_url
        self._base_path = base_path
        self._api_path = api_path

        self._resource_url = "%(base_url)s%(base_path)s%(api_path)s" % (
            {
                'base_url': base_url,
                'base_path': base_path,
                'api_path': api_path
            }
        )

    @staticmethod
    def validate_path(path):
        """Validates the provided path

        :param path: path to validate (string)
        :raise:
            :InvalidUsage: If validation fails.
        """

        if not isinstance(path, str) or not re.match('^/(?:[._a-zA-Z0-9-]/?)+[^/]$', path):
            raise InvalidUsage(
                "Path validation failed - Expected: '/<component>[/component], got: %s" % path
            )

        return True

    @staticmethod
    def get_base_url(use_ssl, instance=None, host=None):
        """Formats the base URL either `host` or `instance`

        :return: Base URL string
        """

        if instance is not None:
            host = ("%s.service-now.com" % instance).rstrip('/')

        if use_ssl is True:
            return "https://%s" % host

        return "http://%s" % host

    def get_appended_custom(self, path_component):
        """Validates the provided path_component, then returns it appended to :prop:`_resource_url`

        :param path_component: Path component string
        :return: Full URL to the custom resource
        """
        self.validate_path(path_component)

        return self._resource_url + path_component

    def get_url(self):
        """Returns :prop:`_resource_url`

        :return: :prop:`_resource_url`
        """
        return self._resource_url


